import os, subprocess
import bionetgen as bng
from distutils import spawn

def find_BNG_path(BNGPATH=None):
    '''
    A simple function finds the path to BNG2.pl from 
    * Environment variable
    * Assuming it's under PATH
    * Given optional path as argument

    Usage: test_bngexec(path)
           test_bngexec()

    Arguments
    ---------
    BNGPATH : str
        (optional) path to the folder that contains BNG2.pl 
    '''
    # TODO: Figure out how to use the BNG2.pl if it's set 
    # in the PATH variable. Solution: set os.environ BNGPATH
    # and make everything use that route 

    # Let's keep up the idea we pull this path from the environment
    if BNGPATH is None:
        try:
            BNGPATH = os.environ["BNGPATH"]
        except:
            pass
    # if still none, try pulling it from cmd line
    if BNGPATH is None:
        bngexec = "BNG2.pl"
        if test_bngexec(bngexec):
            print("BNG2.pl seems to be working")
            # get the source of BNG2.pl
            BNGPATH = spawn.find_executable("BNG2.pl")
            BNGPATH, _ = os.path.split(BNGPATH)
    else:
        bngexec = os.path.join(BNGPATH, "BNG2.pl")
        if test_bngexec(bngexec):
            print("BNG2.pl seems to be working")
        else:
            print("BNG2.pl not working, simulator won't run")
    return BNGPATH, bngexec

def test_bngexec(bngexec):
    '''
    A simple function that test if BNG2.pl given runs

    Usage: test_bngexec(path)

    Arguments
    ---------
    bngexec : str
        path to BNG2.pl to test
    '''
    rc = subprocess.run(["perl", bngexec], stdout=bng.defaults.stdout)
    if rc.returncode == 0:
        return True
    else:
        return False




import platform, os, subprocess
from cement import init_defaults
from cement.utils.version import get_version_banner
from .version import get_version

class BNGDefaults:
    def __init__(self):
        '''
        A class to define the default configuration for cement apps

        Attributes
        ----------
        system : str
            the name of the OS that's running the app
        bng_name : str
            OS appropriate name of the BNG folder
        bng_path : str
            full absolute path to the BNG folder
        lib_path : str
            path to CLI library
        stdout : str
            the name of the subprocess attribute to pass stdout to
        stderr : str
            the name of the subprocess attribute to pass stderr to
        config : dict
            dictionary containing the application defaults
        banner : str
            app banner that gets printed when ran with -v
        '''
        # determine what bng we are using
        system = platform.system() 
        if system == "Linux":
            bng_name = "bng-linux"
        elif system == "Windows":
            bng_name = "bng-win"
        elif system == "Darwin":
            bng_name = "bng-mac"
        # set attributes
        self.system = system
        self.bng_name = bng_name
        # configuration defaults
        CONFIG = init_defaults('bionetgen')
        lib_path = os.path.dirname(__file__)
        lib_path = os.path.split(lib_path)[0]
        CONFIG['bionetgen']['bngpath'] = os.path.join(lib_path, bng_name)
        CONFIG['bionetgen']['notebook'] = {}
        CONFIG['bionetgen']['notebook']["path"] = os.path.join(lib_path, "assets", "bionetgen.ipynb")
        CONFIG['bionetgen']['notebook']["template"] = os.path.join(lib_path, "assets", "bionetgen-temp.ipynb")
        CONFIG["bionetgen"]["notebook"]["name"] = "bng-notebook.ipynb"
        # set attributes
        self.bng_path = os.path.join(lib_path, bng_name)
        self.lib_path = lib_path
        # version banner
        VERSION_BANNER= """
        BioNetGen simple command line interface {}
        {}
        """.format(get_version(), get_version_banner())
        # set attributes
        self.banner = VERSION_BANNER
        # stdout
        CONFIG['bionetgen']['stdout'] = "PIPE"
        CONFIG['bionetgen']['stderr'] = "STDOUT"
        self.stdout = subprocess.PIPE
        self.stderr = subprocess.PIPE
        self.config = CONFIG

defaults = BNGDefaults()



import bionetgen as bng
from bionetgen.main import BioNetGen
import re, functools, subprocess, os, xmltodict, sys, shutil, tempfile
from tempfile import TemporaryDirectory
from tempfile import TemporaryFile
from .utils import find_BNG_path
from .structs import Parameters, Species, MoleculeTypes, Observables, Functions, Compartments, Rules, Actions

# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config['bionetgen']
def_bng_path = conf['bngpath']

###### CORE OBJECT AND PARSING FRONT-END ######
class bngmodel:
    '''
    Main model object and entry point for XMLAPI. The goal of this 
    object is to generate and read the BNGXML of a given BNGL model
    and give the user a pythonic interface to the resulting model object. 

    TODO: * Make model object Cement app configuration aware
    * disentangle XML generation and the model object itself
    * disentangle simulator handling from the model object
    
    Usage: bngmodel(bng_model)
           bngmodel(bng_model, BNGPATH)

    Attributes
    ----------
    active_blocks : list[str]
        a list of the blocks that have been parsed in the model
    _action_list : list[str]
        a list of approved actions
    BNGPATH : str
        path to bionetgen where BNG2.pl lives
    bngexec : str
        path to BNG2.pl
    model_name : str
        name of the model, generally set from the given BNGL file
    recompile : bool
        a tag to keep track if any changes have been made to the model
        via the XML API by the user
    changes : dict
        a list of changes the user have made to the model
    
    Methods
    -------
    parse_model(model_file)
        parses the BNGL model at the given path model_file
    reset_compilation_tags()
        resets compilation tags of each block to keep track of any changes the user
        makes to the model via the API
    generate_xml(model_file, xml_file)
        this generates the BNGXML of the given model file in the xml_file argument
        given
    strip_actions(model_path, folder)
        strips actions from the given model in model_path and writes a model without
        actions in the folder given
    parse_xml(xml_str)
        parses given xml string
    add_action(action_type, action_args)
        adds the action of action_type with arguments given by the optional keyword
        argument action_args which is a list of lists where each element 
        is of the form [ArgumentName, ArgumentValue]
    write_model(model_name)
        write the model in BNGL format to the path given
    write_xml(open_file, xml_type)
        writes the XML of the model into the open_file object given. xml_types allowed
        are BNGXML or SBML formats.
    setup_simulator(sim_type)
        sets up a simulator in bngmodel.simulator where the only current supported 
        type of simulator is libRR for libRoadRunner simulator.
    '''
    def __init__(self, bngl_model, BNGPATH=def_bng_path):
        self.active_blocks = []
        # We want blocks to be printed in the same order
        # every time
        self._action_list = ["generate_network(", "generate_hybrid_model(","simulate(", "simulate_ode(", "simulate_ssa(", "simulate_pla(", "simulate_nf(", "parameter_scan(", "bifurcate(", "readFile(", "writeFile(", "writeModel(", "writeNetwork(", "writeXML(", "writeSBML(", "writeMfile(", "writeMexfile(", "writeMDL(", "visualize(", "setConcentration(", "addConcentration(", "saveConcentration(", "resetConcentrations(", "setParameter(", "saveParameters(", "resetParameters(", "quit(", "setModelName(", "substanceUnits(", "version(", "setOption("]
        self.block_order = ["parameters", "compartments", "moltypes", 
                            "species", "observables", "functions", 
                            "rules", "actions"]
        BNGPATH, bngexec = find_BNG_path(BNGPATH)
        self.BNGPATH = BNGPATH
        self.bngexec = bngexec 
        self.model_name = ""
        self.parse_model(bngl_model)

    @property
    def recompile(self):
        recompile = False
        for block in self.active_blocks:
            recompile = recompile or getattr(self, block)._recompile
        return recompile

    @property
    def changes(self):
        changes = {}
        for block in self.active_blocks:
            changes[block] = getattr(self, block)._changes
        return changes 

    def __str__(self):
        '''
        write the model to str
        '''
        model_str = "begin model\n"
        for block in self.block_order:
            if block in self.active_blocks:
                if block != "actions":
                    model_str += str(getattr(self, block))
        model_str += "\nend model\n"
        if "actions" in self.active_blocks:
            model_str += str(self.actions)
        return model_str

    def __repr__(self):
        return self.model_name

    def __iter__(self):
        active_ordered_blocks = [getattr(self,i) for i in self.block_order if i in self.active_blocks]
        return active_ordered_blocks.__iter__()

    def parse_model(self, model_file):
        # TODO We really need to clean up this method and relevant ones
        # and make it clear what each one of them does. Refactoring time! 
        
        # this route runs BNG2.pl on the bngl and parses
        # the XML instead
        if model_file.endswith(".bngl"):
            print("Attempting to generate XML")
            with TemporaryFile("w+") as xml_file:
                if self.generate_xml(model_file, xml_file):
                    print("Parsing XML")
                    # import IPython;IPython.embed()
                    self.parse_xml(xml_file.read())
                    self.reset_compilation_tags()
                else:
                    print("XML file couldn't be generated")
        elif model_file.endswith(".xml"):
            with open(model_file, "r") as f:
                xml_str = f.read()
                self.parse_xml(xml_str)
            self.reset_compilation_tags()
        else:
            print("The extension of {} is not supported".format(model_file))
            raise NotImplemented

    def reset_compilation_tags(self):
        for block in self.active_blocks:
            getattr(self, block).reset_compilation_tags()

    def generate_xml(self, model_file, xml_file):
        cur_dir = os.getcwd()
        # temporary folder to work in
        with TemporaryDirectory() as temp_folder:
            # make a stripped copy without actions in the folder
            stripped_bngl = self.strip_actions(model_file, temp_folder)
            # run with --xml 
            os.chdir(temp_folder)
            # TODO: take stdout option from app instead
            rc = subprocess.run(["perl",self.bngexec, "--xml", stripped_bngl], stdout=bng.defaults.stdout)
            if rc.returncode == 1:
                print("XML generation failed")
                # go back to our original location
                os.chdir(cur_dir)
                # shutil.rmtree(temp_folder)
                return False
            else:
                # we should now have the XML file 
                path, model_name = os.path.split(stripped_bngl)
                model_name = model_name.replace(".bngl", "")
                written_xml_file = model_name + ".xml"
                # import ipdb;ipdb.set_trace()
                with open(written_xml_file, "r") as f:
                    content = f.read()
                    xml_file.write(content)
                # since this is an open file, to read it later
                # we need to go back to the beginning
                xml_file.seek(0)
                # go back to our original location
                os.chdir(cur_dir)
                return True

    def strip_actions(self, model_path, folder):
        '''
        Strips actions from a BNGL folder and makes a copy
        into the given folder
        '''
        # Get model name and setup path stuff
        path, model_file = os.path.split(model_path)
        # open model and strip actions
        with open(model_path, 'r') as mf:
            # read and strip actions
            mlines = mf.readlines()
            stripped_lines = filter(lambda x: self._not_action(x), mlines)
        # TODO: read stripped lines and store the actions
        # open new file and write just the model
        stripped_model = os.path.join(folder, model_file)
        with open(stripped_model, 'w') as sf:
            sf.writelines(stripped_lines)
        return stripped_model 

    def _not_action(self, line):
        for action in self._action_list:
            if action in line:
                return False
        return True

    def parse_xml(self, xml_str):
        xml_dict = xmltodict.parse(xml_str)
        self.xml_dict = xml_dict
        xml_model = xml_dict['sbml']['model']
        self.model_name = xml_model['@id']
        for listkey in xml_model.keys():
            if listkey == "ListOfParameters":
                param_list = xml_model[listkey]
                if param_list is not None:
                    params = param_list['Parameter']
                    self.parameters = Parameters()
                    self.parameters.parse_xml_block(params)
                    self.active_blocks.append("parameters")
            elif listkey == "ListOfObservables":
                obs_list = xml_model[listkey]
                if obs_list is not None:
                    obs = obs_list['Observable']
                    self.observables = Observables()
                    self.observables.parse_xml_block(obs)
                    self.active_blocks.append("observables")
            elif listkey == "ListOfCompartments":
                comp_list = xml_model[listkey]
                if comp_list is not None:
                    self.compartments = Compartments()
                    comps = comp_list['compartment']
                    self.compartments.parse_xml_block(comps)
                    self.active_blocks.append("compartments")
            elif listkey == "ListOfMoleculeTypes":
                mtypes_list = xml_model[listkey]
                if mtypes_list is not None:
                    mtypes = mtypes_list["MoleculeType"]
                    self.moltypes = MoleculeTypes()
                    self.moltypes.parse_xml_block(mtypes)
                    self.active_blocks.append("moltypes")
            elif listkey == "ListOfSpecies":
                species_list = xml_model[listkey]
                if species_list is not None:
                    species = species_list["Species"]
                    self.species = Species()
                    self.species.parse_xml_block(species)
                    self.active_blocks.append("species")
            elif listkey == "ListOfReactionRules":
                rrules_list = xml_model[listkey]
                if rrules_list is not None:
                    rrules = rrules_list["ReactionRule"]
                    self.rules = Rules()
                    self.rules.parse_xml_block(rrules)
                    self.active_blocks.append("rules")
            elif listkey == "ListOfFunctions":
                # TODO: Optional expression parsing?
                # TODO: Add arguments correctly
                func_list = xml_model[listkey]
                if func_list is not None:
                    self.functions = Functions()
                    funcs = func_list['Function']
                    self.functions.parse_xml_block(funcs)
                    self.active_blocks.append("functions")
        # And that's the end of parsing
        print("XML parsed")

    def add_action(self, action_type, action_args=[]):
        # add actions block and to active list
        if not hasattr(self, "actions"):
            self.actions = Actions()
            self.active_blocks.append("actions")
        self.actions.add_action(action_type, action_args)

    def write_model(self, file_name):
        '''
        write the model to file 
        '''
        model_str = ""
        for block in self.active_blocks:
            model_str += str(getattr(self, block))
        with open(file_name, 'w') as f:
            f.write(model_str)

    def write_xml(self, open_file, xml_type="bngxml"):
        '''
        write new XML to file by calling BNG2.pl again
        '''
        cur_dir = os.getcwd()
        # temporary folder to work in
        with TemporaryDirectory() as temp_folder:
            # write the current model to temp folder
            os.chdir(temp_folder)
            with open("temp.bngl", "w") as f:
                f.write(str(self))
            # run with --xml 
            # TODO: Make output supression an option somewhere
            if xml_type == "bngxml":
                rc = subprocess.run(["perl",self.bngexec, "--xml", "temp.bngl"], stdout=bng.defaults.stdout)
                if rc.returncode == 1:
                    print("XML generation failed")
                    # go back to our original location
                    os.chdir(cur_dir)
                    return False
                else:
                    # we should now have the XML file 
                    with open("temp.xml", "r") as f:
                        content = f.read()
                        open_file.write(content)
                    # go back to beginning
                    open_file.seek(0)
                    os.chdir(cur_dir)
                    return True
            elif xml_type == "sbml":
                rc = subprocess.run(["perl",self.bngexec, "temp.bngl"], stdout=bng.defaults.stdout)
                if rc.returncode == 1:
                    print("SBML generation failed")
                    # go back to our original location
                    os.chdir(cur_dir)
                    return False
                else:
                    # we should now have the SBML file 
                    with open("temp_sbml.xml", "r") as f:
                        content = f.read()
                        open_file.write(content)
                    open_file.seek(0)
                    os.chdir(cur_dir)
                    return True
            else: 
                print("XML type {} not recognized".format(xml_type))
            return False

    def setup_simulator(self, sim_type="libRR"):
        '''
        Sets up a simulator attribute that is a generic front-end
        to all other simulators. At the moment only libroadrunner
        is supported
        '''
        if sim_type == "libRR":
            # we need to add writeSBML action for now
            self.add_action("generate_network", [("overwrite",1)])
            self.add_action("writeSBML", [])
            # temporary file
            with TemporaryFile(mode="w+") as tpath:
                # write the sbml
                self.write_xml(tpath, xml_type="sbml")
                # TODO: Only clear the writeSBML action
                # by adding a mechanism to do so
                self.actions.clear_actions()
                # get the simulator
                self.simulator = bng.sim_getter(model_str=tpath.read(), sim_type=sim_type)
        else:
            print("Sim type {} is not recognized, only libroadrunner \
                   is supported currently by passing libRR to \
                   sim_type keyword argument".format(sim_type))
        return self.simulator

###### CORE OBJECT AND PARSING FRONT-END ######

if __name__ == "__main__":
    # this is to run through a set of 
    # bngl files under the folder validation
    os.chdir("validation")
    bngl_list = os.listdir(os.getcwd())
    bngl_list = filter(lambda x: x.endswith(".bngl"), bngl_list)
    for bngl in bngl_list:
        m = bngmodel(bngl)
        with open('test.bngl', 'w') as f:
            f.write(str(m))
        rc = subprocess.run([m.bngexec, 'test.bngl'])
        if rc.returncode == 1:
            print("issues with the written bngl")
            sys.exit()



from collections import OrderedDict
from .xmlparsers import ObsXML, MolTypeXML, RuleXML, FuncXML, SpeciesXML

###### MODEL STRUCTURES ###### 
# Objects in the model
class ModelBlock:
    '''
    Base block object that will be used for each block in BNGL.

    Attributes
    ----------
    name : str
        Name of the block which will be used to write the BNGL text
    _item_dict : OrderedDict
        every block keeps track of it's items here. the exact nature
        depends on the block itself but this is where every element of
        each block is
    _recompile : boolean
        a tag to keep track if something is changed in the block. this
        will be used in the future to recompile a simulator whne something
        is changed
    _changes : Dict
        a dictionary to keep track of any changes to any items that are in
        a block. this can be used to just keep a history but it can also be
        useful for model validity later

    Methods
    -------
    reset_compilation_tags()
        resets _recompile and _changes tag for the block
    add_item(item_tpl)
        while every block can implement their own add_item method
        the base assumption that each element has a name and value
        so this method takes in (name,value) tuple and sets 
        _item_dict[name] = value
    add_items(item_list)
        loops over every element in the list and uses add_item on it
    print
        prints the block, uses __str__ to get the string
    strip_comment
        removes comments, only used when the blocks are parsed based
        on text BNGL. Not really used anymore
    parse_xml_block(xml)
        while it's not implemented for the base class, each block object
        needs to implement this method where it takes in BNGXML for the 
        block as argument and modifies the object correctly.
    '''
    def __init__(self):
        self.name = "ModelBlock"
        self._item_dict = OrderedDict()
        self._recompile = False
        self._changes = {}

    def __len__(self):
        return len(self._item_dict)

    def __repr__(self):
        # overwrites what the class representation
        # shows the items in the model block in 
        # say ipython
        return str(self._item_dict)

    def __getitem__(self, key):
        if isinstance(key, int):
            # get the item in order
            return list(self._item_dict.keys())[key]
        return self._item_dict[key]

    def __setitem__(self, key, value):
        self._item_dict[key] = value

    def __delitem__(self, key):
        if key in self._item_dict:
            self._item_dict.pop(key)
        else: 
            print("Item {} not found".format(key))

    def __iter__(self):
        return self._item_dict.keys().__iter__()

    def __contains__(self, key):
        return key in self._item_dict
    
    def reset_compilation_tags(self):
        self._recompile = False
        self._changes = {}

    def add_item(self, item_tpl):
        # TODO: try adding evaluation of the parameter here
        # for the future, in case we want people to be able
        # to adjust the math
        # TODO: Error handling, some names will definitely break this
        name, value = item_tpl
        self._item_dict[name] = value
        try:
            setattr(self, name, value)
        except:
            print("can't set {} to {}".format(name, value))
            pass
        self._recompile = True

    def add_items(self, item_list):
        for item in item_list:
            self.add_item(item)

    def print(self):
        print(self)

    def strip_comment(self, line):
        return line[0:line.find("#")]
    
    def parse_xml_block(self, block_xml):
        raise NotImplemented

# TODO: Add a LOT of error handling
class Parameters(ModelBlock):
    '''
    Parameter block object that contains both the expressions and values
    for each parameter defined in the model

    Attributes
    ----------
    expressions : Dict
        this contains the expressions for each item in the block dictionary
    values : dict
        values of each parameter defined in the model
    '''
    def __init__(self):
        self.expressions = {}
        self.values = {}
        super().__init__()
        self.name = "parameters"

    def __setattr__(self, name, value):
        changed = False
        if hasattr(self, "_item_dict"):
            if name in self._item_dict.keys():
                try: 
                    new_value = float(value)
                    changed = True
                    self._item_dict[name] = new_value
                except:
                    self._item_dict[name] = value
        if changed:
            self._changes[name] = new_value
            self.__dict__[name] = new_value
        else:
            self.__dict__[name] = value

    def __str__(self):
        # overwrites what the method returns when 
        # it's converted to string
        block_lines = ["\nbegin {}".format(self.name)]
        for item in self._item_dict.keys():
            block_lines.append("  " + "{} {}".format(item, self._item_dict[item]))
        block_lines.append("end {}\n".format(self.name))
        return "\n".join(block_lines)

    def add_item(self, item_tpl):
        name, value = item_tpl
        self._item_dict[name] = value
        try:
            setattr(self, name, value)
        except:
            print("can't set {} to {}".format(name, value))
            pass
        self._recompile = True

    def parse_xml_block(self, block_xml):
        # 
        if isinstance(block_xml, list):
            for b in block_xml:
                self.values[b['@id']] = b['@value']
                if '@expr' in b:
                    self.expressions[b['@id']] = b['@expr']
                    self.add_item((b['@id'],b['@expr']))
                else:
                    self.add_item((b['@id'],b['@value']))
        else:
            self.values[block_xml['@id']] = block_xml['@value']
            if '@expr' in block_xml:
                self.expressions[block_xml['@id']] = block_xml['@expr']
                self.add_item((block_xml['@id'], block_xml['@expr']))
            else:
                self.add_item((block_xml['@id'], block_xml['@value']))
        # 

class Species(ModelBlock):
    '''
    Species block object that contains starting species values defined
    in the model

    Item dictionary contains the Species object as the name and the 
    concentration as the value so outside of parsing this is basically 
    the same as the base object.
    '''
    def __init__(self):
        super().__init__()
        self.name = "species"

    def __str__(self):
        # overwrites what the method returns when 
        # it's converted to string
        block_lines = ["\nbegin {}".format(self.name)]
        for item in self._item_dict.keys():
            block_lines.append("  " + "{} {}".format(item,self._item_dict[item]))
        block_lines.append("end {}\n".format(self.name))
        return "\n".join(block_lines)

    def __getitem__(self, key):
        if isinstance(key, str):
            # our keys are objects
            for ikey in self._item_dict:
                if key == str(ikey):
                    return self._item_dict[ikey]
        if isinstance(key, int):
            # get the item in order
            return list(self._item_dict.keys())[key]
        return self._item_dict[key]

    def __setitem__(self, key, value):
        if isinstance(key, str):
            for ikey in self._item_dict:
                if key == str(ikey):
                    self._item_dict[ikey] = value
            return
        if isinstance(key, int):
            k = list(self._item_dict.keys())[key]
            self._item_dict[k] = value
            return
        self._recompile = True
        self._item_dict[key] = value

    def __contains__(self, key):
        for ikey in self._item_dict:
            if key == str(ikey):
                return True
        return False

    def parse_xml_block(self, block_xml):
        if isinstance(block_xml, list):
            for sd in block_xml:
                xmlobj = SpeciesXML(sd)
                self.add_item((xmlobj,sd['@concentration']))
        else:
            xmlobj = SpeciesXML(block_xml)
            self.add_item((xmlobj,block_xml['@concentration']))

    def add_item(self, item_tpl):
        # TODO is this necessary? this should be defined in base
        name, val = item_tpl
        self._item_dict[name] = val

class MoleculeTypes(ModelBlock):
    '''
    Molecule type block object that contains molecule types defined in the 
    model. 

    Item dictionary contains MolTypeXML object as the name has an empty 
    string as the value thus outside of parsing this is basically 
    the same as the base object.
    '''
    def __init__(self):
        super().__init__()
        self.name = "molecule types"

    def __repr__(self):
        return str(list(self._item_dict.keys()))

    def add_item(self, item_tpl):
        name, = item_tpl
        self._item_dict[name] = ""
        self._recompile = True

    def __getitem__(self, key):
        if isinstance(key, str):
            # our keys are objects
            for ikey in self._item_dict:
                if key == str(ikey):
                    return self._item_dict[ikey]
        if isinstance(key, int):
            # get the item in order
            return list(self._item_dict.keys())[key]
        return self._item_dict[key]

    def __setitem__(self, key, value):
        for ikey in self._item_dict:
            if key == str(ikey):
                self._recompile = True
                self._item_dict[ikey] = value

    def __contains__(self, key):
        for ikey in self._item_dict:
            if key == str(ikey):
                return True
        return False

    def __str__(self):
        # overwrites what the method returns when 
        # it's converted to string
        block_lines = ["\nbegin {}".format(self.name)]
        for item in self._item_dict.keys():
            block_lines.append("  " + "{}".format(item))
        block_lines.append("end {}\n".format(self.name))
        return "\n".join(block_lines)

    def parse_xml_block(self, block_xml):
        if isinstance(block_xml, list):
            for md in block_xml:
                xmlobj = MolTypeXML(md)
                self.add_item((xmlobj,))
        else:
            xmlobj = MolTypeXML(block_xml)
            self.add_item((xmlobj,))


class Observables(ModelBlock):
    '''
    Observable block object that contains observables defined in the 
    model. 

    Item dictionary contains the name of the observable as the name
    and a list of the form [ObservableType, ObservableObject] as value. 
    '''
    def __init__(self):
        super().__init__()
        self.name = "observables"

    def __setattr__(self, name, value):
        if hasattr(self, "_item_dict"):
            if name in self._item_dict.keys():
                self._recompile = True
                self._changes[name] = value
                self._item_dict[name][1] = value
        self.__dict__[name] = value

    def add_item(self, item_tpl): 
        otype, name, obj = item_tpl
        self._item_dict[name] = [otype, obj]
        try:
            setattr(self, name, obj)
        except:
            print("can't set {} to {}".format(name, obj))
            pass
        self._recompile = True

    def __str__(self):
        # overwrites what the method returns when 
        # it's converted to string
        block_lines = ["\nbegin {}".format(self.name)]
        for item in self._item_dict.keys():
            block_lines.append("  " + 
                    "{} {} {}".format(self._item_dict[item][0],
                                      item,
                                      self._item_dict[item][1]))
        block_lines.append("end {}\n".format(self.name))
        return "\n".join(block_lines)

    def __getitem__(self, key):
        if isinstance(key, int):
            # get the item in order
            return self._item_dict[list(self._item_dict.keys())[key]][1]
        return self._item_dict[key]

    def parse_xml_block(self, block_xml):
        #
        if isinstance(block_xml, list):
            for b in block_xml:
                xmlobj = ObsXML(b['ListOfPatterns'])
                self.add_item((b['@type'], b['@name'], xmlobj))
        else: 
            xmlobj = ObsXML(block_xml['ListOfPatterns'])
            self.add_item((block_xml['@type'], block_xml['@name'], xmlobj))
        # 


class Functions(ModelBlock):
    '''
    Function block object that contains functions defined in the 
    model. 

    Item dictionary contains the name of the function as the name
    and function expression as value.
    '''
    def __init__(self):
        super().__init__()
        self.name = "functions"

    # TODO: Fix this such that we can re-write functions
    def __str__(self):
        # overwrites what the method returns when 
        # it's converted to string
        block_lines = ["\nbegin {}".format(self.name)]
        for item in self._item_dict.keys():
            block_lines.append("  " + 
                    "{} = {}".format(item, self._item_dict[item]))
        block_lines.append("end {}\n".format(self.name))
        return "\n".join(block_lines)

    def parse_xml_block(self, block_xml):
        if isinstance(block_xml, list):
             for func in block_xml:
                 xmlobj = FuncXML(func)
                 self.add_item(xmlobj.item_tuple)
        else:
             xmlobj = FuncXML(block_xml)
             self.add_item(xmlobj.item_tuple)

class Compartments(ModelBlock):
    '''
    Compartments block object that contains compartments defined in the 
    model. 

    Item dictionary contains the name of the compartment as the name
    and a list of the form [Dimensionality, Size, ParentCompartment] 
    as the value.
    '''
    def __init__(self):
        super().__init__()
        self.name = "compartments"

    def __str__(self):
        # overwrites what the method returns when 
        # it's converted to string
        block_lines = ["\nbegin {}".format(self.name)]
        for item in self._item_dict.keys():
            comp_line = "  {} {} {}".format(item, 
                            self._item_dict[item][0],
                            self._item_dict[item][1])
            if self._item_dict[item][2] is not None:
                comp_line += " {}".format(self._item_dict[item][2])
            block_lines.append(comp_line)
        block_lines.append("end {}\n".format(self.name))
        return "\n".join(block_lines)

    def add_item(self, item_tpl):
        name, dim, size, outside = item_tpl
        self._recompile = True
        self._item_dict[name] = [dim, size, outside]

    def parse_xml_block(self, block_xml):
        # 
        if isinstance(block_xml, list):
            for comp in block_xml:
                cname = comp['@id']
                dim = comp['@spatialDimensions']
                size = comp['@size']
                if '@outside' in comp:
                    outside = comp['@outside']
                else:
                    outside = None
                self.add_item( (cname, dim, size, outside) )
        else:
            cname = block_xml['@id']
            dim = block_xml['@spatialDimensions']
            size = block_xml['@size']
            if '@outside' in block_xml:
                outside = block_xml['@outside']
            else:
                outside = None
            self.add_item( (cname, dim, size, outside) )
        #

class Rules(ModelBlock):
    '''
    Compartments block object that contains compartments defined in the 
    model. 

    Item dictionary contains the name of the compartment as the name
    and a list of the form [Dimensionality, Size, ParentCompartment] 
    as the value.
    '''
    def __init__(self):
        super().__init__()
        self.name = "reaction rules"

    def __str__(self):
        block_lines = ["\nbegin {}".format(self.name)]
        for item in self._item_dict.keys():
            block_lines.append(str(self._item_dict[item]))
        block_lines.append("end {}\n".format(self.name))
        return "\n".join(block_lines)

    def __iter__(self):
        return self._item_dict.values().__iter__()

    def parse_xml_block(self, block_xml):
        if isinstance(block_xml, list):
            for rd in block_xml:
                xmlobj = RuleXML(rd)
                self.add_item((xmlobj.name, xmlobj))
        else:
            xmlobj = RuleXML(block_xml)
            self.add_item((xmlobj.name, xmlobj))
        self.consolidate_rules()

    def consolidate_rules(self):
        '''
        Generated XML only has unidirectional rules
        and uses "_reverse_" tag to make bidirectional 
        rules for NFSim. Take all the "_reverse_" tagged
        rules and convert them to bidirectional rules
        '''
        delete_list = []
        for item_key in self._item_dict:
            rxn_obj  = self._item_dict[item_key]
            if item_key.startswith("_reverse_"):
                # this is the reverse of another reaction
                reverse_of = item_key.replace("_reverse_", "")
                # ensure we have the original
                if reverse_of in self._item_dict:
                    # make bidirectional and add rate law
                    r1 = self._item_dict[reverse_of].rate_constants[0]
                    r2 = rxn_obj.rate_constants[0]
                    self._item_dict[reverse_of].set_rate_constants((r1,r2))
                    # mark reverse for deletion
                    delete_list.append(item_key)
        # delete items marked for deletion
        for del_item in delete_list:
            self._item_dict.pop(del_item)

class Actions(ModelBlock):
    '''
    Action block object that contains actions defined in the 
    model. This is the one object that doesn't need a begin/end
    block tag to be in the model. 

    Item dictionary contains the action type as the name and a list
    of arguments for that action as value. Each argument is of form 
    [ArgumentName, ArgumentValue], method argument value is written
    with quotes since that's the only one that needs quotes in BNGL.

    Attributes
    ----------
    _action_list : list[str]
        list of available action names
    '''
    def __init__(self):
        super().__init__()
        self.name = "actions"
        self._action_list = ["generate_network", "generate_hybrid_model",
            "simulate", "simulate_ode", "simulate_ssa", "simulate_pla", 
            "simulate_nf", "parameter_scan", "bifurcate", "readFile", 
            "writeFile", "writeModel", "writeNetwork", "writeXML", 
            "writeSBML", "writeMfile", "writeMexfile", "writeMDL", 
            "visualize", "setConcentration", "addConcentration", 
            "saveConcentration", "resetConcentrations", "setParameter", 
            "saveParameters", "resetParameters", "quit", "setModelName", 
            "substanceUnits", "version", "setOption"]

    def __str__(self):
        # TODO: figure out every argument that has special 
        # requirements, e.g. method requires the value to 
        # be a string
        block_lines = []
        for item in self._item_dict.keys():
            action_str = "{}(".format(item) + "{"
            for iarg,arg in enumerate(self._item_dict[item]):
                val = arg[1]
                arg = arg[0]
                if iarg > 0:
                    action_str += ","
                if arg == "method":
                    action_str += '{}=>"{}"'.format(arg, val)
                else:
                    action_str += '{}=>{}'.format(arg, val)
            action_str += "})"
            block_lines.append(action_str)
        return "\n".join(block_lines)

    def add_action(self, action_type, action_args):
        '''
        adds action, needs type as string and args as list of tuples
        (which preserve order) of (argument, value) pairs
        '''
        if action_type in self._action_list:
            self._item_dict[action_type] = action_args
        else:
            print("Action type {} not valid".format(action_type))

    def clear_actions(self):
        self._item_dict.clear()
###### MODEL STRUCTURES ###### 



# All classes that deal with patterns 
class Pattern:
    '''
    Pattern object. Fundamentally it's just a list of molecules
    which are defined later. 

    Attributes
    ----------
    _bonds : Bonds
        setting a pattern requires you to keep track of all bonds to
        correctly label them, this object tracks everything
    compartment : str
        compartment of the overall pattern (not the same thing as 
        molecule compartment, those have their own)
    label : str
        label of the overall pattern (not the same thing as molecule 
        label, those have their own)
    molecules : list[Molecule]
        list of molecule objects that are in the pattern

    Methods
    -------
    _parse_xml(xml)
        parses the entire pattern xml
    _process_mol(mol_xml)
        parses the molecule xml to add to pattern molecule list
    _process_comp(comp_xml)
        parses component xml to add to the molecule
    '''
    def __init__(self, pattern_xml):
        self._bonds = Bonds()
        self._compartment = None
        self._label = None
        self.molecules = []
        # sets self.molecules up 
        self._parse_xml(pattern_xml)

    @property
    def compartment(self):
        return self._compartment 

    @compartment.setter
    def compartment(self, value):
        # TODO: Build in logic to set the 
        # outer compartment
        # print("Warning: Logical checks are not complete")
        self._compartment = value
        # by default, once the outer compartment is set
        # we will set the compartment of each molecule
        # to that new compartment. 
        for molec in self.molecules:
            molec.compartment = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # TODO: Build in logic to set 
        # the outer label
        # print("Warning: Logical checks are not complete")
        self._label = value

    def __str__(self):
        sstr = ""
        for imol, mol in enumerate(self.molecules):
            if imol == 0 and self.compartment is not None:
                sstr += "@{}:".format(self.compartment)
            if imol == 0 and self.label is not None:
                sstr += "%{}:".format(self.label)
            if imol > 0:
                sstr += "."
            sstr += str(mol)
        return sstr

    def __repr__(self):
        return str(self)

    def __getitem__(self, key):
        return self.molecules[key]

    def __iter__(self):
        return self.molecules.__iter__()

    # TODO: Implement __contains__

    def _parse_xml(self, xml):
        if '@compartment' in xml:
            self.compartment = xml['@compartment']
        if "@label" in xml:
            self.label = xml["@label"]
        if "ListOfBonds" in xml:
            self._bonds.set_xml(xml["ListOfBonds"]["Bond"])
        mols = xml['ListOfMolecules']['Molecule']
        if isinstance(mols, list):
            # list of molecules
            for imol, mol in enumerate(mols):
                mol_obj = self._process_mol(mol)
                self.molecules.append(mol_obj)
        else:
            # a single molecule
            mol_obj = self._process_mol(mols)
            self.molecules.append(mol_obj)

    def _process_mol(self, mol_xml):
        # we are going to store molecules, components
        # and compartments in a separate dictionary 
        # for use later
        mol_obj = Molecule()
        mol_obj.name = mol_xml['@name'] 
        if "@label" in mol_xml:
            mol_obj.label = mol_xml["@label"]
        if "ListOfComponents" in mol_xml:
            # Single molecule can't have bonds
            mol_obj.components = self._process_comp(mol_xml["ListOfComponents"]["Component"])
        if '@compartment' in mol_xml:
            mol_obj.compartment = mol_xml['@compartment']
        return mol_obj

    def _process_comp(self, comp_xml):
        # bonds = compartment id, bond id 
        # comp xml can be a list or a dict
        comp_list = []
        if isinstance(comp_xml, list):
            # we have multiple and this is a list
            for icomp, comp in enumerate(comp_xml):
                comp_obj = Component()
                comp_obj.name = comp['@name']
                if "@label" in comp:
                    comp_obj.label = comp['@label']
                if "@state" in comp:
                    comp_obj.state = comp['@state']
                if comp["@numberOfBonds"] != '0':
                    bond_id = self._bonds.get_bond_id(comp)
                    for bi in bond_id:
                        comp_obj.bonds.append(bi)
                comp_list.append(comp_obj)
        else:
            # single comp, this is a dict
            comp_obj = Component()
            comp_obj.name = comp_xml['@name']
            if "@label" in comp_xml:
                comp_obj.label = comp_xml['@label']
            if "@state" in comp_xml:
                comp_obj.state = comp_xml['@state']
            if comp_xml['@numberOfBonds'] != '0':
                bond_id = self._bonds.get_bond_id(comp_xml)
                for bi in bond_id:
                    comp_obj.bonds.append(bi)
            comp_list.append(comp_obj)
        return comp_list

class Molecule:
    '''
    Pattern object. Fundamentally it's just a list of molecules
    which are defined later. 

    Attributes
    ----------
    name : str
        name of the molecule
    compartment : str
        compartment of the molecule
    label : str
        label of the molecule
    components : list[Component]
        list of components for this molecule

    Methods
    -------
    add_component(name, state=None, states=[])
        add a component object to the list of components with name 
        "name", current state "state" or a list of states 
        (for molecule types) "states"
    '''
    def __init__(self):
        self._name = "0"
        self._components = []
        self._compartment = None
        self._label = None

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.components[key]

    def __iter__(self):
        return self.components.__iter__()

    # TODO: implement __setitem__,  __contains__

    def __str__(self):
        mol_str = self.name
        if self.label is not None:
            mol_str += "%{}".format(self.label)
        if len(self.components) > 0:
            mol_str += "("
            for icomp, comp in enumerate(self.components):
                if icomp > 0:
                    mol_str += ","
                mol_str += str(comp)
            mol_str += ")"
        if self.compartment is not None:
            mol_str += "@{}".format(self.compartment)
        return mol_str
    
    ### PROPERTIES ### 
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        # print("Warning: Logical checks are not complete")
        # TODO: Check for invalid characters
        self._name = value

    @property
    def components(self):
        return self._components

    @components.setter
    def components(self, value):
        # print("Warning: Logical checks are not complete")
        self._components = value

    def __repr__(self):
        return str(self)

    @property
    def compartment(self):
        return self._compartment

    @compartment.setter
    def compartment(self, value):
        # print("Warning: Logical checks are not complete")
        self._compartment = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # print("Warning: Logical checks are not complete")
        self._label = value

    def _add_component(self, name, state=None, states=[]):
        comp_obj = Component()
        comp_obj.name = name
        comp_obj.state = state
        comp_obj.states = states
        self.components.append(comp_obj)

    def add_component(self, name, state=None, states=[]):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._add_component(name, state, states)

class Component:
    '''
    Component object that describes the state, label and bonding
    for each component.

    Attributes
    ----------
    name : str
        name of the component
    label : str
        label of the component
    state : str
        state of the component, not used for molecule types
    states : list[str]
        list of states for molecule types
    bonds : list[Bond]
        list of bond objects that describes bonding of the component

    Methods
    -------
    add_state()
        not implemented. will eventually be used to add additional states
        to an existing component
    add_bond()
        not implemented. will eventually be used to add additional bonds
        to an existing component  
    '''
    def __init__(self):
        self._name = ""
        self._label = None
        self._state = None
        self._states = []
        self._bonds = []

    def __repr__(self):
        return str(self)

    def __str__(self):
        comp_str = self.name
        # only for moltypes
        if len(self.states) > 0:
            for istate, state in enumerate(self.states):
                comp_str += "~{}".format(state)
        # for any other pattern
        if self.state is not None:
            comp_str += "~{}".format(self.state)
        if self.label is not None:
            comp_str += "%{}".format(self.label)
        if len(self.bonds) > 0:
            for bond in self.bonds:
                comp_str += "!{}".format(bond)
        return comp_str

    ### PROPERTIES ### 
    @property
    def name(self):
        return self._name
   
    @name.setter
    def name(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._name = value

    @property
    def label(self):
        return self._label
   
    @label.setter
    def label(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._label = value

    @property
    def state(self):
        return self._state
   
    @state.setter
    def state(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._state = value

    @property
    def states(self):
        return self._states
   
    @states.setter
    def states(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._states = value

    @property
    def bonds(self):
        return self._bonds
   
    @bonds.setter
    def bonds(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._bonds = value

    def _add_state(self):
        raise NotImplemented

    def add_state(self):
        self._add_state()

    def _add_bond(self):
        raise NotImplemented

    def add_bond(self):
        self._add_bond()

###### BONDS #####
class Bonds:
    def __init__(self, bonds_xml=None):
        self.bonds_dict = {}
        if bonds_xml is not None:
            self.resolve_xml(bonds_xml)

    def set_xml(self, bonds_xml):
        self.resolve_xml(bonds_xml)

    def get_bond_id(self, comp):
        # Get the ID of the bond from an XML id something 
        # belongs to, e.g. O1_P1_M1_C2 
        num_bonds = comp["@numberOfBonds"]
        comp_id = comp["@id"]
        try: 
            num_bond = int(num_bonds)
        except: 
            # This means we have something like +/?
            return num_bonds
        # use the comp_id to find the bond index from 
        # self.bonds_dict 
        comp_key = self.get_tpl_from_id(comp_id)
        bond_id = self.bonds_dict[comp_key]
        return bond_id
        
    def get_tpl_from_id(self, id_str):
        # ID str is looking like O1_P1_M2_C3
        # we are going to assume a 4-tuple per key
        id_list = id_str.split("_")
        id_tpl = tuple(id_list)
        return id_tpl

    def tpls_from_bond(self, bond):
        s1 = bond["@site1"] 
        s2 = bond["@site2"]
        id_list_1 = s1.split("_")
        s1_tpl = tuple(id_list_1)
        id_list_2 = s2.split("_")
        s2_tpl = tuple(id_list_2)
        return (s1_tpl, s2_tpl) 

    def resolve_xml(self, bonds_xml):
        # self.bonds_dict is a dictionary you can key
        # with the tuple taken from the ID and then 
        # get a bond ID cleanly
        if isinstance(bonds_xml, list):
            for ibond, bond in enumerate(bonds_xml): 
                bond_partner_1, bond_partner_2 = self.tpls_from_bond(bond)
                if bond_partner_1 not in self.bonds_dict:
                    self.bonds_dict[bond_partner_1] = [ibond+1]
                else:
                    self.bonds_dict[bond_partner_1].append([ibond+1])
                if bond_partner_2 not in self.bonds_dict:
                    self.bonds_dict[bond_partner_2] = [ibond+1]
                else:
                    self.bonds_dict[bond_partner_2].append(ibond+1)
        else:
            bond_partner_1, bond_partner_2 = self.tpls_from_bond(bonds_xml)
            self.bonds_dict[bond_partner_1] = [1]
            self.bonds_dict[bond_partner_2] = [1]
###### BONDS #####



from .model import bngmodel 
from .structs import Parameters, Species, MoleculeTypes, Observables, Functions,Compartments, Rules
from .pattern import Pattern, Molecule, Bonds
from .xmlparsers import ObsXML, MolTypeXML, RuleXML, FuncXML, SpeciesXML


from .pattern import Pattern, Molecule, Component, Bonds

###### XMLObjs ###### 
class XMLObj:
    '''
    Base object that contains XMLs that is the parent of
    all XML object that will be used for BNGL blocks as
    we read from BNGXML. This sets up the python internals
    such as __repr__ and __str__ for the rest of the XML 
    classes.

    This base class assumes at least two methods will be
    defined by each class that inherits this base object
    which are defined in the methods below. 

    Attributes
    ----------
    xml : ??
        XML string to be parsed for the block

    Methods
    -------
    resolve_xml(xml)
        the method that parses the XML associated with
        the instance and adjust is appropriately 
    gen_string()
        the method to generate the string from the information
        contained in the instance
    '''
    def __init__(self, xml):
        self.xml = xml
        self.resolve_xml(self.xml)

    def __repr__(self):
        return self.gen_string()

    def __str__(self):
        return self.gen_string()

class ObsXML(XMLObj):
    '''
    Observable XML object. Observables are a list of 
    patterns where a pattern is a list of molecules. 

    Attributes
    ----------
    patterns : list
        list of Pattern objects that make up the observable
    '''
    def __init__(self, xml):
        self.patterns = []
        super().__init__(xml)

    def __iter__(self):
        return self.patterns.__iter__()

    def __getitem__(self, key):
        return self.patterns[key]

    def gen_string(self):
        obs_str = ""
        for ipat, pat in enumerate(self.patterns):
            if ipat > 0:
                obs_str += ","
            obs_str += str(pat)
        return obs_str

    def resolve_xml(self, obs_xml):
        patterns = obs_xml['Pattern']
        if isinstance(patterns, list):
            # we have multiple patterns so this is a list
            for ipattern, pattern in enumerate(patterns): 
                # 
                self.patterns.append(Pattern(pattern))
        else:
            self.patterns.append(Pattern(patterns))

class SpeciesXML(Pattern):
    '''
    Species XML object. Species are a list of molecules. 

    Attributes
    ----------
    molecules : list
        list of molecules objects that make up the species
    '''
    def __init__(self, xml):
        self._xml = xml
        self._bonds = Bonds()
        self._label = None
        self._compartment = None
        self.molecules = []
        # sets self.molecules up 
        self._parse_xml(xml)

class MolTypeXML(XMLObj):
    '''
    Molecule Type XML object. Molecules types are like molecules
    but with multiple states for each component. 

    Attributes
    ----------
    molecule : Molecule
        a molecule object that forms this molecule type

    Methods
    -------
    add_component(name, states)
        adds component with name "name" and optional states given by 
        a list of strings
    '''
    def __init__(self, xml):
        super().__init__(xml)

    def add_component(self, name, states=None):
        self.molecule.add_component(name, states=states)

    def gen_string(self):
        return str(self.molecule)

    def resolve_xml(self, molt_xml):
        mol_obj = Molecule()
        mol_obj.name = molt_xml['@id'] 
        if 'ListOfComponentTypes' in molt_xml:
            comp_obj = Component()
            comp_dict = molt_xml['ListOfComponentTypes']['ComponentType']
            if '@id' in comp_dict:
                comp_obj.name = comp_dict["@id"]
                if "ListOfAllowedStates" in comp_dict:
                    # we have states
                    al_states = comp_dict['ListOfAllowedStates']['AllowedState']
                    if isinstance(al_states, list):
                        for istate, state in enumerate(al_states):
                            comp_obj.states.append(state["@id"])
                    else:
                        comp_obj.states.append(state["@id"])
                mol_obj.components.append(comp_obj)
            else:
                # multiple components
                for icomp, comp in enumerate(comp_dict):
                    comp_obj = Component()
                    comp_obj.name = comp['@id']
                    if "ListOfAllowedStates" in comp:
                        # we have states
                        al_states = comp['ListOfAllowedStates']['AllowedState']
                        if isinstance(al_states, list):
                            for istate, state in enumerate(al_states):
                                comp_obj.states.append(state['@id'])
                        else:
                            comp_obj.states.append(al_states['@id'])
                    mol_obj.components.append(comp_obj)
        self.molecule = mol_obj

class RuleXML(XMLObj):
    '''
    Molecule Type XML object. Molecules types are like molecules
    but with multiple states for each component. 

        A rule is a tuple (list of reactant patterns, list of 
    product patterns, list of rate constant functions)

    Attributes
    ----------
    bidirectoinal : boolean
        list of molecules objects that make up the species
    name : str
        name of the rule, if exists
    reactants : list[Pattern]
        list of patterns that form the reactant side of the rule
    products : list[Pattern]
        List of patterns that form the products side of the rule
    rate_constants : list[Str]
        list of 1 or 2 rate constants

    Methods
    -------
    side_string(name, states)
        gets the string for a side of the reaction, given a list of pattern objects
    set_rate_constants(rate_cts)
        takes an iterable of 1 or 2 rate constants and adjust the object
        and sets bidirectionality
    resolve_rate_law(rate_xml)
        parses XML for rate constants and adds it to the object
    resolve_rxn_side(side_xml)
        parses the XML for either reactants or products and adds it to 
        the object
    '''
    def __init__(self, pattern_xml):
        self.bidirectional = False
        super().__init__(pattern_xml)

    def __iter__(self):
        return self.iter_tpl.__iter__()

    def gen_string(self):
        if self.bidirectional:
            return "{}: {} <-> {} {},{}".format(self.name, self.side_string(self.reactants), self.side_string(self.products), self.rate_constants[0], self.rate_constants[1])
        else:
            return "{}: {} -> {} {}".format(self.name, self.side_string(self.reactants), self.side_string(self.products), self.rate_constants[0])

    def side_string(self, patterns):
        side_str = ""
        for ipat, pat in enumerate(patterns):
            if ipat > 0:
                side_str += " + "
            side_str += str(pat)
        return side_str

    def set_rate_constants(self, rate_cts):
        if len(rate_cts) == 1:
            self.rate_constants = [rate_cts[0]]
            self.bidirectional = False
        elif len(rate_cts) == 2: 
            self.rate_constants = [rate_cts[0], rate_cts[1]]
            self.bidirectional = True
        else:
            print("1 or 2 rate constants allowed")
    
    def resolve_xml(self, pattern_xml):
        # 
        rule_name = pattern_xml['@name']
        self.name = rule_name
        self.reactants = self.resolve_rxn_side(pattern_xml['ListOfReactantPatterns'])
        self.products = self.resolve_rxn_side(pattern_xml['ListOfProductPatterns'])
        if 'RateLaw' not in pattern_xml:
            print("Rule seems to be missing a rate law, please make sure that XML exporter of BNGL supports whatever you are doing!")
        self.rate_constants = [self.resolve_ratelaw(pattern_xml['RateLaw'])]
        self.rule_tpl = (self.reactants, self.products, self.rate_constants)

    def resolve_ratelaw(self, rate_xml):
        rate_type = rate_xml['@type']
        if rate_type == 'Ele':
            rate_cts_xml = rate_xml['ListOfRateConstants']
            rate_cts = rate_cts_xml['RateConstant']['@value']
        elif rate_type == 'Function':
            rate_cts = rate_xml['@name']
        elif rate_type == 'MM' or rate_type == "Sat":
            # A function type 
            rate_cts = rate_type + "("
            args = rate_xml['ListOfRateConstants']["RateConstant"]
            if isinstance(args, list):
                for iarg, arg in enumerate(args):
                    if iarg > 0:
                        rate_cts += ","
                    rate_cts += arg["@value"]
            else:
                rate_cts += args["@value"]
            rate_cts += ")"
        else:
            print("don't recognize rate law type")
        return rate_cts

    def resolve_rxn_side(self, side_xml):
        # this is either reactant or product
        if side_xml is None:
            return [Molecule()]
        elif 'ReactantPattern' in side_xml:
            # this is a lhs/reactant side
            sl = []
            side = side_xml['ReactantPattern']
            if '@compartment' in side:
                self.react_comp = side['@compartment']
            else:
                self.react_comp = None
            if isinstance(side, list):
                # this is a list of reactant patterns
                for ireact, react in enumerate(side):
                    sl.append(Pattern(react))
            else: 
                sl.append(Pattern(side))
            return sl
        elif "ProductPattern" in side_xml:
            # rhs/product side
            side = side_xml['ProductPattern']
            sl = []
            if '@compartment' in side:
                self.prod_comp = side['@compartment']
            else:
                self.prod_comp = None
            if isinstance(side, list):
                # this is a list of product patterns
                for iprod, prod in enumerate(side):
                    sl.append(Pattern(prod))
            else: 
                sl.append(Pattern(side))
            return sl
        else: 
            print("Can't parse rule XML {}".format(side_xml))

class FuncXML(XMLObj):
    # TODO for some reason the resolve_xml doesn't set the full
    # function string and it's also not used downstream. 

    '''
    Function XML object. Functions are expressions and IDs which 
    is the name of the function with potential arguments. 

    Attributes
    ----------
    item_tuple
        tuple of (function_str, expression) where function_str is the 
        string that forms the name + arguments of the function (e.g. g(x))
        and expression is the definition of the function.
    
    Methods
    -------
    get_arguments(arg_xml)
        given the XML of arguments, this methods pulls out the list of 
        arguments the function needs and returns a list of strings. 
    '''
    def __init__(self, pattern_xml):
        super().__init__(pattern_xml)

    def resolve_xml(self, func_xml):
        fname = func_xml['@id']
        expression = func_xml['Expression']
        args = []
        if 'ListOfArguments' in func_xml:
            args = self.get_arguments(func_xml['ListOfArguments']['Argument'])
        expr = func_xml['Expression']
        func_str = fname + "("
        if len(args) > 0:
            for iarg, arg in enumerate(args):
                if iarg > 0:
                    func_str += ","
                func_str += arg
        func_str += ")"
        self.item_tuple = (func_str, expression)
        full_str = "{} = {}".format(func_str, expression)
        return full_str 

    def get_arguments(self, arg_xml):
        args = []
        if isinstance(arg_xml, list):
            for arg in arg_xml:
                args.append(arg['@id'])
            return args
        else:
            return [arg_xml['@id']]
###### PATTERNS ###### 



from setuptools import setup, find_packages
from bionetgen.core.version import get_version
import itertools as itt

VERSION = get_version()

#### BNG DOWNLOAD START ####
# Handle BNG download and inclusion
import platform,json
import urllib.request
# let's pull URLs for each distribution
# in the latest distribution
rls_url = "https://api.github.com/repos/RuleWorld/bionetgen/releases/latest"
rls_resp = urllib.request.urlopen(rls_url)
rls_json_txt = rls_resp.read()
rls_json = json.loads(rls_json_txt)
assets = rls_json['assets']
for asset in assets:
    browser_url = asset['browser_download_url']
    if 'linux' in browser_url:
        linux_url = browser_url
    elif 'mac' in browser_url:
        mac_url = browser_url
    elif 'win' in browser_url and "tgz" in browser_url:
        windows_url = browser_url

# next set the correct URL for platform
# system = platform.system() 
# if system == "Linux":
#     bng_url = linux_url
# elif system == "Windows":
#     bng_url = windows_url
# elif system == "Darwin":
#     bng_url = mac_url
# else:
#     print("Setup doesn't know your system! {} \
#             BioNetGen won't be installed".format(system))
#     bng_url = None

# next download and place the appropriate files
import os,shutil,tarfile,zipfile
bng_downloaded = False
for iurl,bng_url in enumerate([linux_url, mac_url, windows_url]):
    # this lists the stuff in bng distribution 
    # to move in our python package
    if iurl < 2:
        to_move = [["BNG2.pl"], ["bin","NFsim"], ["bin", "run_network"], ["bin","sundials-config"], ["Perl2"], ["VERSION"]]
    elif iurl == 2:
        to_move = [["BNG2.pl"], ["bin","NFsim.exe"], ["bin", "run_network.exe"], ["bin","sundials-config"], ["Perl2"], ["VERSION"], ["bin","cyggcc_s-seh-1.dll"],["bin","cygstdc++-6.dll"],["bin","cygwin1.dll"],["bin","cygz.dll"],["bin","libgcc_s_dw2-1.dll"],["bin","libstdc++-6.dll"]]
    # import file and download libraries
    ext = bng_url.split(".")[-1]
    fname = "bng.{}".format(ext)
    # download bng distro
    urllib.request.urlretrieve(bng_url, fname)
    # file extension dependent actions, tar for 
    # mac and linux, zip for windows
    if iurl < 2:
    # if "tgz" == ext or "gz" == etx:
        bng_arch = tarfile.open(fname)
        fold_name = bng_arch.getnames()[0]
        bng_arch.extractall()
        # make sure bionetgen/bng exists
        if iurl == 0:
            bng_path_to_move = "bionetgen/bng-linux"
        elif iurl == 1:
            bng_path_to_move = "bionetgen/bng-mac"

        if os.path.isdir(bng_path_to_move):
            shutil.rmtree(bng_path_to_move)
        os.mkdir(bng_path_to_move)
        os.mkdir(os.path.join(bng_path_to_move, "bin"))
        # move items in
        for item in to_move:
            item_list = [fold_name] + item
            item_path = list(itt.accumulate(item_list, lambda x,y: os.path.join(x,y)))[-1]
            to_move_item = [bng_path_to_move] + item
            to_move_path = list(itt.accumulate(to_move_item, lambda x,y: os.path.join(x,y)))[-1]
            shutil.move(item_path, to_move_path)
        # we got bionetgen in
        bng_downloaded = True
        # Done unpacking, remove useless files
        bng_arch.close()
        os.remove(fname)
        shutil.rmtree(fold_name)
    if iurl == 2:
    # elif "zip" == etx:
        # TODO: handle zip/windows case
        # bng_arch = zipfile.Zipfile(fname)
        # fold_name = bng_arch.namelist()[0]
        # bng_arch.extractall()
        bng_arch = tarfile.open(fname)
        fold_name = bng_arch.getnames()[0]
        bng_arch.extractall()
        # bng folder
        if iurl == 2:
            bng_path_to_move = "bionetgen/bng-win"

        if os.path.isdir(bng_path_to_move):
            shutil.rmtree(bng_path_to_move)
        os.mkdir(bng_path_to_move)
        os.mkdir(os.path.join(bng_path_to_move, "bin"))
        # move items in
        for item in to_move:
            item_list = [fold_name] + item
            item_path = list(itt.accumulate(item_list, lambda x,y: os.path.join(x,y)))[-1]
            to_move_item = [bng_path_to_move] + item[:-1]
            to_move_path = list(itt.accumulate(to_move_item, lambda x,y: os.path.join(x,y)))[-1]
            shutil.move(item_path, to_move_path)
        # we got bionetgen in
        bng_downloaded = True
        # Done unpacking, remove useless files
        bng_arch.close()
        os.remove(fname)
        shutil.rmtree(fold_name)

# if bng_downloaded:
#     # TODO: only add if not there
#     with open("MANIFEST.in", "a") as f:
#         f.write("recursive-include bionetgen/bng-linux *\n")
#         f.write("recursive-include bionetgen/bng-mac *\n")
#         f.write("recursive-include bionetgen/bng-win *\n")
#### BNG DOWNLOAD DONE ####

f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='bionetgen',
    version=VERSION,
    description='A simple CLI for BioNetGen ',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Ali Sinan Saglam',
    author_email='als251@pitt.edu',
    url='https://github.com/ASinanSaglam/BNG_cli',
    license='unlicensed',
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    package_data={'bionetgen': ['bng*/*', 'assets/*']},
    zip_safe=False,
    include_package_data=True,
    entry_points="""
        [console_scripts]
        bionetgen = bionetgen.main:main
    """,
    install_requires=[
        "cement",
        "nbopen",
        "numpy",
        "pyyaml",
        "colorlog",
        "xmltodict",
    ]
)



import os 
import numpy as np

class BNGResult:
    '''
    Class that loads in gdat/cdat/scan files
    
    Usage: BNGResult(path="/path/to/folder") OR
           BNGResult(direct_path="/path/to/file.gdat") 

    Arguments
    ---------
    path : str
        path that points to a folder containing files to be
        loaded by the class
    direct_path : str
        path that directly points to a file to load

    Methods
    -------
    load(fpath)
        loads in the direct path to the file and returns
        numpy.recarray
    '''
    def __init__(self, path=None, direct_path=None):
        # TODO Make it so that with path you can supply an
        # extension or a list of extensions to load in
        self.results = {}
        self.names = {} 
        if direct_path is not None:
            path, fname = os.path.split(direct_path)
            fnoext, fext = os.path.splitext(fname)
            self.direct_path = direct_path
            self.file_name = fnoext
            self.file_extension = fext
            self.names[fnoext] = direct_path
            self.results[fnoext] = self.load(direct_path)
        elif path is not None:
            # TODO change this pattern so that each method 
            # is stand alone and usable. 
            self.path = path
            self.find_gdat_files() 
            self.load_results()
        else:
            print("BNGResult needs either a path or a direct path kwarg to load gdat/cdat/scan files from")
            os.sys.exit()

    def load(self, fpath):
        path, fname = os.path.split(fpath)
        fnoext, fext = os.path.splitext(fname)
        if fext == ".gdat" or fext == ".cdat":
            return self._load_dat(fpath)
        elif fext == ".scan": 
            return self._load_scan(fpath)
        else:
            print("BNGResult doesn't know the file type of {}".format(fpath))
            return None

    def _load_scan(self, fpath):
        return self._load_dat(fpath)

    def find_gdat_files(self):
        files = os.listdir(self.path)
        gdat_files = filter(lambda x: x.endswith(".gdat"), files)
        for gdat_file in gdat_files:
            name = gdat_file.replace(".gdat", "")
            self.names[name] = gdat_file

    def load_results(self):
        for name in self.names: 
            gdat_path = os.path.join(self.path, self.names[name])
            self.results[name] = self._load_dat(gdat_path)

    def _load_dat(self, path, dformat="f8"):
        '''
        This function takes a path to a gdat/cdat file as a string and loads that 
        file into a numpy structured array, including the correct header info.
        TODO: Add link
    
        Optional argument allows you to set the data type for every column. See
        numpy dtype/data type strings for what's allowed. TODO: Add link
        '''
        # First step is to read the header, 
        # we gotta open the file and pull that line in
        f = open(path)
        header = f.readline()
        f.close()
        # Ensure the header info is actually there
        assert header.startswith("#"), "No header line that starts with #"
        # Now turn it into a list of names for our struct array
        header = header.replace("#","")
        headers = header.split()
        # For a magical reason this is how numpy.loadtxt wants it, 
        # in tuples passed as a dictionary with names/formats as keys
        names = tuple(headers)
        formats = tuple([dformat for i in range(len(headers))])
        # return the loadtxt result as a record array
        # which is similar to pandas data format without the helper functions
        return np.rec.array(np.loadtxt(path, dtype={'names': names, 'formats': formats}))



import cement, os, platform, subprocess
import bionetgen as bng
from cement.core.exc import CaughtSignal
from .core.exc import BioNetGenError
from .core.main import runCLI
from .core.main import plotDAT 
from .core.notebook import BNGNotebook

# pull defaults defined in core/defaults
CONFIG = bng.defaults.config
VERSION_BANNER = bng.defaults.banner

class BNGBase(cement.Controller):
    '''
    Base cement controller for BioNetGen CLI

    Used to set meta attributes like program name (label) as well 
    as command line arguments. Each method is a subcommand in the 
    command line with it's own command line arguments. 

    Subcommands
    -------
    run
        runs a model given by -i in folder given by -o
    notebook
        generates and opens a notebook for a model given by -i, optional
    plot
        plots a gdat/cdat/scan file given by -i into file supplied by -o
    '''

    class Meta:
        label = "bionetgen"
        description = "A simple CLI to bionetgen <https://bionetgen.org>. Note that you need Perl installed."
        help = "bionetgen"
        arguments = [
                # TODO figure out a good solution for when bngpath is set from config file AND CLI
                # until then we'll disable the CLI argument
                # (['-bp','--bngpath'],dict(type=str,
                #                           default=CONFIG['bionetgen']['bngpath'],
                #                           help="Optional path to BioNetGen folder you want the CLI to use")),
                # TODO: Auto-load in BioNetGen version here
                (['-v','--version'],dict(action="version",
                                         version=VERSION_BANNER)),
                # (['-s','--sedml'],dict(type=str,
                #                        default=CONFIG['bionetgen']['bngpath'],
                #                        help="Optional path to SED-ML file, if available the simulation \
                #                              protocol described in SED-ML will be ran")),
        ]

    # This overwrites the default behavior and runs the CLI object from core/main
    # which in turn just calls BNG2.pl with the supplied options
    @cement.ex(
            help="Runs a given model using BNG2.pl",
            arguments=[
                (["-i","--input"],{"help":"Path to BNGL file (required)",
                                   "default": None,
                                   "type": str,
                                   "required": True}),
                (["-o","--output"],{"help":"Optional path to output folder (default: \".\")",
                                    "default": ".",
                                    "type": str})
            ]
    )
    def run(self):
        '''
        This is the main run functionality of the CLI. 
        
        It uses a convenience function defined in core/main 
        to run BNG2.pl using subprocess, given the set of arguments 
        in the command line and the configuraions set by the defaults 
        as well as the end-user. 
        '''
        args = self.app.pargs
        runCLI(self.app.config, args)

    @cement.ex(
            help="Starts a Jupyter notebook to help run and analyze \
                  bionetgen models",
            arguments=[
                (["-i","--input"],{"help":"Path to BNGL file to use with notebook",
                                   "default": None,
                                   "type": str,
                                   "required": False}),
                (["-o","--output"],{"help":"(optional) File to write the notebook in",
                                   "default": "",
                                   "type": str}),
                (["-op","--open"],{"help":"(optional) If given, the notebook will by opened using nbopen",
                                   "action": 'store_true'}),
            ]
    )
    def notebook(self):
        '''
        Notebook subcommand that boots up a Jupyter notebook using the 
        nbopen library. It uses a BNGNotebook class defined in core/notebook. 

        The default template can be found under assets and in the future
        will likely be replaced by a standard templating tool (e.g. Jinja).
        
        The default base template is agnostic to the model and if -i is given
        the template then will be adjusted to load in the model supplied.
        '''
        args = self.app.pargs
        if args.input is not None:
            # we want to use the template to write a custom notebok
            notebook = BNGNotebook(CONFIG["bionetgen"]["notebook"]["template"], INPUT_ARG=args.input)
        else:
            # just use the basic notebook
            notebook = BNGNotebook(CONFIG["bionetgen"]["notebook"]["path"])
        # find our file name
        if len(args.output) == 0:
            fname = CONFIG["bionetgen"]["notebook"]["name"]
        else:
            fname = args.output
        # write the notebook out
        notebook.write(fname)
        # open the notebook with nbopen
        stdout = getattr(subprocess, CONFIG["bionetgen"]["stdout"])
        stderr = getattr(subprocess, CONFIG["bionetgen"]["stderr"])
        if args.open:
            rc = subprocess.run(["nbopen", fname], stdout=stdout, stderr=stderr)

    @cement.ex(
            help="Rudimentary plotting of gdat/cdat/scan files",
            arguments=[
                (["-i","--input"],{"help":"Path to .gdat/.cdat file to use plot",
                                   "default": None,
                                   "type": str,
                                   "required": True}),
                (["-o","--output"],{"help":"Optional path for the plot (default: \"$model_name.png\")",
                                    "default": ".",
                                    "type": str}),
                (["--legend"],{"help":"To plot the legend or not (default: False)",
                                   "default": False,
                                   "action": "store_true",
                                   "required": False}),
                (["--xmin"],{"help":"x-axis minimum (default: determined from data)",
                                   "default": None}),
                (["--xmax"],{"help":"x-axis maximum (default: determined from data)",
                                   "default": False}),
                (["--ymin"],{"help":"y-axis minimum (default: determined from data)",
                                   "default": False}),
                (["--ymax"],{"help":"y-axis maximum (default: determined from data)",
                                   "default": False}),
                (["--xlabel"],{"help":"x-axis label (default: time)",
                                   "default": False}),
                (["--ylabel"],{"help": "y-axis label (default: concentration)",
                                   "default": False}),
                (["--title"],{"help": "title of plot (default: determined from input file)",
                                   "default": False})
            ]
    )
    def plot(self):
        """
        Plotting subcommand for very basic plotting using a convenience function
        defined in core/main.

        Currently we support gdat/cdat/scan file plotting, in a very basic manner. 
        This command expects a space separated file where each column is a series. 
        The first column is used for the x-axis and the rest is used as y-axis 
        and every series is plotted. 

        See bionetgen plot -h for all the allowed options.  
        """
        args = self.app.pargs
        # we need to have gdat/cdat files
        assert args.input.endswith(".gdat") or args.input.endswith(".cdat") or args.input.endswith(".scan"), "Input file has to be either a gdat or a cdat file"
        plotDAT(args.input, args.output, kw=dict(args._get_kwargs()))

class BioNetGen(cement.App):
    '''
    Cement app for BioNetGen CLI

    Used to set configuration options like config default,
    exiting on close and setting log handler. Currently set
    attributes are below.

    Attributes
    ----------
    label : str
        name of the application
    config_defaults : str
        the default set of configuration options, set in BNGDefaults object
    config_handler: str
        the name of the config handler, determines the syntax of the config files
    config_file_suffix: str
        the suffix to be used for config files
    config_files: list of str
        additional list of config files to enable
    exit_on_close : boolean
        determine if the app should exit when the key function is ran
    extensions : list of str
        extensions to be used with cement framework
    log_handler: str
        name of the log handler
    handlers: list of obj
        list of objects derived from cement.Controller that handles the actual CLI
    '''

    class Meta:
        label = 'bionetgen'

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            'yaml',
            'colorlog',
        ]

        # configuration handler
        config_handler = 'configparser'

        # configuration file suffix
        config_file_suffix = '.conf'

        # add current folder to the list of config dirs
        config_files = [
            './.{}.conf'.format(label)
        ]

        # set the log handler
        log_handler = 'colorlog'

        # register handlers
        handlers = [
            BNGBase
        ]

class BioNetGenTest(cement.TestApp,BioNetGen):
    """ 
    A sub-class of BioNetGen CLI application for testing 
    purposes. See tests/test_bionetgen.py for examples.
    """

    class Meta:
        label = "bionetgen"


def main():
    with BioNetGen() as app:
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except BioNetGenError as e:
            print('BioNetGenError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()
